/*
    \brief Deklarationen der Funktionen für die Listen der Zutrittskontrolle Version 2.

    \author Bernd Ottmann
    \date   21.02.2003
*/
#pragma once
#ifndef DATAFOX_ACCESS_CONTROL_LISTS_H
#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define DATAFOX_ACCESS_CONTROL_LISTS_H
#endif

#include "df_defines.h"

#if (DF_API_VERSION >= 0x04010100)

/*!
    \brief Bibliotheksinterne Datenpuffer für Listendaten der Zutrittskontrolle löschen und initalisieren.

    Die in der Bibliothek bereitgestellten Datenpuffer für die zu übertragenenden Listendaten werden gelöscht und
    für eine Verwendung mittels DFCMakeEntrance2List() und DFCLoadEntrance2List() initalisiert.

    \ingroup access_control_v2

    \export{_DFCClearEntrance2ListBuffer@8, DFCClearEntrance2ListBuffer}

    \param[in] connectionNumber Zu verwendende Verbindungsnummer.
    \param[in] listNumber Listennummer. Bei Angabe von -1 werden alle vorhandenen geloescht.
    <table class="paramvalues">
        <tr><th>Wert</th><th>Listenname</th></tr>
        <tr><td>0</td><td>Reader</td></tr>
        <tr><td>1</td><td>Identification</td></tr>
        <tr><td>2</td><td>Location</td></tr>
        <tr><td>3</td><td>Time</td></tr>
        <tr><td>4</td><td>Holiday</td></tr>
        <tr><td>5</td><td>Events</td></tr>
        <tr><td>6</td><td>Action</td></tr>
        <tr><td>7</td><td>Action2</td></tr>
    </table>

    \since 04.01.01 enthalten
*/
DF_API_EXTERN_DECORATE void DF_API_CALL DFCClearEntrance2ListBuffer( int connectionNumber, int listNumber );

/*!
    \brief Kopieren eines aufbereiteten Datenarrays für Listendaten der Zutrittskontrolle in den Biblioteksinternen Datenpuffer.

    Das zu übergebende Datenarray wird in den Bibliotheksinternen Datenpuffer kopiert und für eine spätere
    Übertragung mittels DFCLoadEntrance2List() zwischengespeichert.

    \ingroup access_control_v2

    \export{_DFCMakeEntrance2List@24, DFCMakeEntrance2List}

    \param[in] connectionNumber Zu verwendende Verbindungsnummer.
    \param[in] listNumber Listennummer.
    <table class="paramvalues">
        <tr><th>Wert</th><th>Listenname</th><th>Größe pro Datensatz in Byte</th></tr>
        <tr><td>0</td><td>Reader</td><td>33</td></tr>
        <tr><td>1</td><td>Identification</td><td>64</td></tr>
        <tr><td>2</td><td>Location</td><td>20</td></tr>
        <tr><td>3</td><td>Time</td><td>25</td></tr>
        <tr><td>4</td><td>Holiday</td><td>21</td></tr>
        <tr><td>5</td><td>Events</td><td>17</td></tr>
        <tr><td>6</td><td>Action</td><td>21</td></tr>
        <tr><td>7</td><td>Action2</td><td>39</td></tr>
    </table>
    \param[in] recordCount Anzahl der Datensätze im übergebenen Datenarray.
    \param[in] size Anzahl Bytes des Datenarray.\n Brechnet sich aus (\a recordCount * Größe pro Datensatz).
    \param[in] buffer Datenarray mit den Daten.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.

    \since 04.01.01 enthalten
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCMakeEntrance2List( int connectionNumber, int listNumber, int recordCount, int size,
                                                            const unsigned char *buffer, int *errorNumber );

/*!
    \brief Übertragen der bibliotheksinternen Datenpuffer zum Gerät.

    Es werden die mit DFCMakeEntrance2List() importierten Listendaten in das Gerät geschrieben. Vorhandene Listendaten
    werden hierbei erst nach einer erfolgreichen Übernahme überschrieben.

    \ingroup access_control_v2

    \export{_DFCLoadEntrance2List@16, DFCLoadEntrance2List}

    \attention Bitte beachten Sie die [Hinweise zur Lebensdauer](\ref dfc_lists_memory) der eingesetzten Flash-Bausteine.

    \note Die Listen besitzen teilweise interne Bezüge zueinander. Um die Datenkonsistenz zu wahren müssen Sie sicherstellen,
    das bei hinzufügen neuer Datensätze die sich über mehrere Listen auswirken diese in einem Bundle übertragen werden. \n
    Importieren Sie hierzu bitte alle zusammengehörenden Listen mittels DFCMakeEntrance2List und überragen diese dann mit
    mit dem Wert -1 für den Parameter number. Es wird dadurch sichergestellt, dass alle Listen erst übernommen werden,
    wenn die Überragung durchweg erfolgreich war.

    \param[in] connectionNumber Zu verwendende Verbindungsnummer.
    \param[in] deviceAddress Anzusprechende Geräteadresse.
    \param[in] listNumber Listennummer. Bei Angabe von -1 werden alle importierten Listendaten übertragen.
    <table class="paramvalues">
        <tr><th>Wert</th><th>Listenname</th></tr>
        <tr><td>0</td><td>Reader</td></tr>
        <tr><td>1</td><td>Identification</td></tr>
        <tr><td>2</td><td>Location</td></tr>
        <tr><td>3</td><td>Time</td></tr>
        <tr><td>4</td><td>Holiday</td></tr>
        <tr><td>5</td><td>Events</td></tr>
        <tr><td>6</td><td>Action</td></tr>
        <tr><td>7</td><td>Action2</td></tr>
        <tr><td>-1</td><td>Alle vorliegenden Listendaten des bibliotheksinternen Puffers</td></tr>
    </table>
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.

    \since 04.01.01 enthalten
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCLoadEntrance2List( int connectionNumber, int deviceAddress, int listNumber, int *errorNumber );

#endif

#endif
