// file_studio_setup.cpp: Implementierung der Klasse FileAES.
//
//////////////////////////////////////////////////////////////////////

#include "file/file_aes.h"
#include <sys/stat.h>
#include <time.h>
#include <algorithm>

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

FileAES::FileAES()
    : m_flash_header( *reinterpret_cast<StructFlashHeader *>(m_eep_mem.data()) )
    , m_eep( m_eep_mem.data() )
    , m_file_header( *reinterpret_cast<StructFileStudioHeader *>(m_file_header_mem.data()) )
    , m_config( *reinterpret_cast<StructConfig *>(m_config_mem.data()) )
    , m_flash( m_flash_mem.data() )
    , m_extended_header( *reinterpret_cast<StructExtendedFlash *>(m_extended_mem.data()) )
    , m_flashExtended( m_extended_mem.data() )
{
    // Initialisieren der 32bit-CRC Umgebung.
    InitCRC32Table();

    // Dokument aufsetzen
    SetDefault();
}

FileAES::~FileAES()
{
}

void FileAES::SetDefault( int32_t flags )
{
    // Eeprom 
    m_eep_mem.fill( 0xFF );

    // Header auf Standard setzen.
    resetFileHeader();

    // Zusaetlicher Setupspeicher
    m_flashSizeSU2 = FLASH_SIZE_SU2;
    // Flashbereiche
    m_flash_mem.fill( 0xFF );
    // Konfigbereich
    m_config_mem.fill( 0 );
    // Extended Flash wird jedoch mit 0 iniatlisiert
    m_extended_mem.fill( 0 );

    // Der erweiterte Speicherbereich wird mit einer dynamischen Struktur begonnen,
    // deren erster Parameter ein Check - UINT32 - Wert enthaelt.

    // Alles mit Null initialisieren.
    memset( &m_extended_header, 0, sizeof( StructExtendedFlash ) );
    // Struktur fuer erweiterten Bereich aufsetzen
    m_extended_header.check = EXTENDED_CHECK;
    m_extended_header.size = sizeof( StructExtendedFlash );
    m_extended_header.offsetFreeScope = m_extended_header.size;
    // Flag fuer Verarbeitung
    m_extended_header.flags = EXTF_NEW;

    // Die Listenangaben werden bei gesetzten flag 1 beibehalten?
    if ( (flags & 0x01) != 0x01 )
    {
        // Listenpfade löschen
        for ( auto &s : m_listFileName )
        {
            s.clear();
        }
        for ( auto &s : m_listEntranceFileName )
        {
            s.clear();
        }
        for ( auto &s : m_listEntrance2FileName )
        {
            s.clear();
        }
        for ( auto &s : m_strExtendedFilePath )
        {
            s.clear();
        }
    }
}

// Setzt den Bereich des Dateiheaders zurück und trägt die grundlegenden Werte ein.
void FileAES::resetFileHeader()
{
    // Dateiheader
    m_file_header_mem.fill( 0 );

    // Dateikopf aufsetzen
    m_file_header.check = magic_header_number();
    m_file_header.bugfixSize = sizeof_StructFileStudioHeader_1;
    m_file_header.size = sizeof( StructFileStudioHeader );

    // Ggf. vorher gesetzte Werte wieder eintragen.
    m_file_header.studioVersion = header_version( m_studioVersion.master(), m_studioVersion.major(), m_studioVersion.minor(), m_studioVersion.build() );
    
    m_flash_header.studioVersion = m_studioVersion;
    m_flash_header.libraryVersion = m_libraryVersion;

    // Es wird mindestens diese Version benoetigt.
    // Die Version wird ggf. auf die benoetigte in der jeweiligen MakeImage-Funktion angepasst.
    NeedStudioVersion( "04.02.02.08" );
    NeedLibraryVersion( "04.02.02.08" );
}

/*
return:
0: Fehler beim Erstellen der Headerdaten eingetreten.
1: Dateiheaderdaten erfolgreich erstellt und eingetragen.
*/
int32_t FileAES::buildHeader()
{
    // CRC der Dateibereiche
    m_file_header.crcEep = GetCrcEep();
    m_file_header.crcEepExtended = 0; // War in Version 1, CRC von Restheaderbereich berechnen und eintragen
    m_file_header.crcFlash = GetCrcFlash();
    m_file_header.crcListFile = GetCrcListFile();
    m_file_header.crcConfig = GetCrcConfig();
    m_file_header.crcFlashExtended = GetCrcFlashExtended();
    m_file_header.crcListEntranceFile = GetCrcListEntranceFile();
    m_file_header.crcListEntrance2File = GetCrcListEntrance2File();
    m_file_header.crcListExtendedFile = GetCrcListExtendedFile();

    // Size und CRC der Ersten Version immer eintragen.
    m_file_header.bugfixSize = sizeof_StructFileStudioHeader_1;
    m_file_header.bugfixCRC = GetCRC32( m_file_header_mem.data(), m_file_header.bugfixSize - sizeof( uint32_t ) );

    // Size dieser Version eintragen, wenn noch nicht aktuell.
    if ( m_file_header.size < sizeof( StructFileStudioHeader ) )
        m_file_header.size = sizeof( StructFileStudioHeader );

    // Gesamt-CRC fuer Header berechnen und eintragen.
    *(uint32_t *)&m_file_header_mem[m_file_header.size - sizeof( uint32_t )] = GetCRC32( m_file_header_mem.data(), m_file_header.size - sizeof( uint32_t ) );

    // Wenn eine aktueller Version als hier aktuell vorhanden geladen wurde:
    // Die Groessenangabe und Daten werden beibehalten.
    // Es kann davon ausgegangen werden, dass im vorangegangenen Ladeverlauf die Mindestversion des Studio oder Library geprueft wurde.
    // Wenn das Laden erfolgreich durchgefuehrt werden konnte handelt es sich um hinzugefuegte Daten die nicht so relevant sind oder ggf. gar nicht verwendet werden.
    // Es wird in einem solchen Fall mit der Struktur gearbeitet wie sie ist.

    return 1;
}

/*
return:
0: Laengenangabe oder Pruefsumme des Headers falsch.
1: Pruefungen erfolgreich und ohne Fehler durchgefuehrt.
*/
int32_t FileAES::checkHeader()
{
    if ( IsHeader() )
    {
        // Version 1, Angaben sind bei allen Versionen verfuegbar.
        // Laengenangabe liegt im moeglichen Bereich?
        if ( m_file_header.bugfixSize != sizeof_StructFileStudioHeader_1 )
        {
            return 0;
        }

        // Pruefen ob die Headerdaten ggf. korrupt sind?
        if ( *(uint32_t *)&m_file_header_mem[m_file_header.bugfixSize - sizeof( uint32_t )] != GetCRC32( m_file_header_mem.data(), m_file_header.bugfixSize - sizeof( uint32_t ) ) )
        {
            return 0;
        }


        // Version >= 2
        // Laengenangabe liegt im moeglichen Bereich?
        if ( m_file_header.size != 0 && (m_file_header.size < sizeof_StructFileStudioHeader_2 || m_file_header.size > m_file_header_mem.size()) )
        {
            return 0;
        }
        else if ( m_file_header.size != 0 )
        {
            // Pruefen ob die Headerdaten ggf. korrupt sind?
            if ( *(uint32_t *)&m_file_header_mem[m_file_header.size - sizeof( uint32_t )] != GetCRC32( m_file_header_mem.data(), m_file_header.size - sizeof( uint32_t ) ) )
            {
                return 0;
            }

            // CRC loeschen, da geprueft und sich ggf. mit Werten dieser Stukturversion ueberlagert.
            *(uint32_t *)&m_file_header_mem[m_file_header.size - sizeof( uint32_t )] = 0;
        }
        else // size == 0
        {
            // Dahinterliegende Daten als ungueltig behandeln.
            memset( &m_file_header_mem[m_file_header.bugfixSize], 0, m_file_header_mem.size() - m_file_header.bugfixSize );
        }

        // Version 1
        // CRC loeschen, da geprueft.
        *(uint32_t *)&m_file_header_mem[m_file_header.bugfixSize - sizeof( uint32_t )] = 0;

        // Ist Version 2?
        //if ( m_file_header.size == sizeof_StructFileStudioHeader_2 )

        // Wenn eine aktueller Version als hier aktuell vorhanden geladen wuerde:
        // Es kann davon ausgegangen werden, dass im nachfolgenden Ladeverlauf die Mindestversion des Studio oder Library nicht erreicht wird,
        // wenn nicht, dann handelt es sich um hinzugefuegte Daten die nicht so relevant sind oder ggf. gar nicht verwendet werden.
        // Es wird in einem solchen Fall mit der Struktur geargeitet wie sie ist.
    }

    return 1;
}

/**
 * @brief Lesen eines Strings, mit führender Längenangabe, aus einer Datei.
 *
 * @param[out] s Gelesener String.
 * @param[in] fp Datei aus der ab der aktuellen Position gelesen wird.
 *
 * @returns Anzahl gelesener Bytes inkl. Bytes für die Längenangabe.
 */
uint32_t FileAES::readCString( std::string &s, std::FILE *fp )
{
    // String nach MFC - CString Konvention lesen.
    uint32_t result = 0;
    bool fail = false;
    uint32_t size = 0;
    uint8_t len[2];
    s.clear();

    // Stringlänge ermitteln. Die Längenangabe liegt entweder mit 1 Byte vor dem String solange
    // sie kleiner als 255 ist, ansonsten folgt einem Byte mit dem Wert 255 ein 16bit Wert.
    result = static_cast<uint32_t>(std::fread( len, sizeof( uint8_t ), 1, fp )); // ein Byte Länge
    if ( result == sizeof( uint8_t ) )
    {
        if ( len[0] == 0xFF )
        {
            result += static_cast<uint32_t>(std::fread( len, sizeof( uint8_t ), 2, fp )); // zwei Byte Länge
            if ( result == 3 * sizeof( uint8_t ) )
            {
                size = *reinterpret_cast<uint16_t *>(len);
            }
            else
            {
                // Fehler beim Ermitteln der Länge.
                fail = true;
            }
        }
        else
        {
            size = len[0];
        }

        if ( !fail )
        {
            // String für die zu lesende Anzahl Zeichen anpassen.
            s.resize( size, 0 );
        }
    }
    else
    {
        // Fehler beim Ermitteln der Länge.
        fail = true;
    }

    if ( !fail )
    {
        if ( (size = static_cast<uint32_t>(std::fread( &s[0], sizeof( uint8_t ), s.size(), fp ))) != s.size() )
        {
            // String an die wirklich gelesene Anzahl anpassen.
            s.resize( size, 0 );
        }

        // Gelesene Länge anpassen.
        result += size;
    }

    return result;
}

/**
 * @brief Schreiben eines Strings, mit führender Längenangabe, in eine Datei.
 *
 * @param[in] s Zu schreibender String.
 * @param[in] fp Datei in die ab der aktuellen Position geschrieben wird.
 *
 * @returns Anzahl geschriebener Bytes inkl. Bytes für die Längenangabe.
 */
uint32_t FileAES::writeCString( const std::string &s, std::FILE *fp )
{
    // String nach MFC - CString Konvention schreiben.
    uint32_t result = 0;
    uint32_t size;
    bool fail = false;

    if ( s.size() > 0xFF )
    {
        uint8_t len[3];
        len[0] = 0xFF; // Markierung das eine 16bit Laengenangabe folgt.
        *reinterpret_cast<uint16_t *>(&len[1]) = static_cast<uint16_t>(s.size());
        result = static_cast<uint32_t>(std::fwrite( len, sizeof( uint8_t ), sizeof( len ), fp ));
        if ( result != sizeof( len ) )
        {
            fail = true;
        }
    }
    else
    {
        uint8_t len = static_cast<uint8_t>(s.size());
        result = static_cast<uint32_t>(std::fwrite( &len, sizeof( uint8_t ), sizeof( len ), fp ));
        if ( result != sizeof( len ) )
        {
            fail = true;
        }
    }

    // String rausschreiben
    if ( !fail )
    {
        if ( (size = static_cast<uint32_t>(std::fwrite( s.c_str(), sizeof( uint8_t ), s.size(), fp ))) != s.size() )
        {
            // Fehler beim Schreiben des gesamten Strings.
        }

        // Geschriebene Länge anpassen.
        result += size;
    }

    return result;
}

/**
 * @brief Laden der Setupdaten aus der angegebenen Datei.
 *
 * @param[in] file_path Angabe der zu ladenden Datei.
 * @param[out] error Erhält bei Fehler die entsprechende Fehlernummer.
                     <table class="paramvalues">
                         <tr><th>Wert</th><th>Bedeutung</th></tr>
                         <tr><td>-1</td><td> </td></tr>
                         <tr><td>-2</td><td>Öffnen der Datei gescheitert. </td></tr>
                         <tr><td>-3</td><td>Lesen aus der Datei gescheitert. </td></tr>
                         <tr><td>-4</td><td>Prüfsummenfehler im Dateiinformationblock. </td></tr>
                         <tr><td>-5</td><td>Zum Öffnen der Setupdatei wird mindestens das DatafoxStudioIV - Version
                                            m_file.m_file_header.minStudioVersion benötigt.
                                            In der Setupdatei werden Einstellungen und Erweiterungen verwendet,
                                            die durch dieses DatafoxStudioIV noch nicht unterstützt werden.
                                            Bitte aktualisiseren Sie die Anwendung auf die angegebene Mindestversion.
                                    </td></tr>
                        <tr><td>-6</td><td>Warnung: Um alle in der Setupdatei vorliegenden Informationen zu Unterstützen
                                           wird mindestens das DatafoxStudioIV - Version m_file.m_file_header.shouldStudioVersion
                                           benötigt. Ein Abspeichern der Setupdatei mit dem vorliegenden DatafoxStudioIV
                                           führt zu einem Verlust von Verwaltungsspezifischen Dateidaten.
                        </td></tr>
                        <tr><td>-7</td><td>Zum Öffnen der Setupdatei wird mindestens die DFComDLL - Version m_file.m_file_header.minLibraryVersion
                                           benötigt. In der Setupdatei werden Einstellungen und Erweiterungen verwendet,
                                           die durch diese DFComDLL noch nicht unterstützt werden. Bitte aktualisiseren
                                           Sie die Bibliothek auf die angegebene Mindestversion.
                        </td></tr>
                        <tr><td>-8</td><td>Warnung: Um alle in der Setupdatei vorliegenden Informationen zu Unterstützen
                                           wird mindestens die DFComDLL - Version m_file.m_file_header.shouldLibraryVersion
                                           benötigt. Ein Abspeichern der Setupdatei mit dem vorliegenden DatafoxStudioIV
                                           führt zu einem Verlust von Verwaltungsspezifischen Dateidaten.
                        </td></tr>
                        <tr><td>-9</td><td>Unerwartetes Setupdateiformat.</td></tr>
                        <tr><td>-10</td><td> </td></tr>
                        <tr><td>-11</td><td> </td></tr>
                        <tr><td>-12</td><td> </td></tr>
                        <tr><td>-13</td><td> </td></tr>
                        <tr><td>-14</td><td> </td></tr>
                        <tr><td>-15</td><td> </td></tr>
                        <tr><td>-16</td><td> </td></tr>
                        <tr><td>-17</td><td> </td></tr>
                        <tr><td>-18</td><td>Prüfsummenfehler im Dateninhalt.</td></tr>
                     </table>

 * @param[in] flags Zusätzliche Angabe über durchzuführende Prüfungen während dem Ladevorgang.
                     <table class="paramvalues">
                         <tr><th>Wert</th><th>Bedeutung</th></tr>
                         <tr><td>1</td><td>Ignoriere die benötigte Mindestversionsangabe des DatafoxStudioIV.</td></tr>
                         <tr><td>2</td><td>Ignoriere die Angabe der sollte-Version des DatafoxStudioIV.</td></tr>
                         <tr><td>4</td><td>Ignoriere die benötigte Mindestversionsangabe des DatafoxLibraryIV.</td></tr>
                         <tr><td>8</td><td>Ignoriere die Angabe der sollte-Version des DatafoxLibraryIV.</td></tr>
                     </table>
 *
 * @retval 0 Fehler bei Ausführung.
 * @retval 1 Ausführung war erfolgreich.
 */
int32_t FileAES::load( const std::string &file_path, int32_t *error, Enum_flag_ignore_version flags )
{
    int32_t res = 1;
    struct stat st;
    uint32_t sizeSetup;
    uint32_t listFileSize = 0;
    uint32_t entranceListFileSize = 0;
    uint32_t entrance2ListFileSize = 0;
    uint32_t extendedFilePathSize = 0;

    if ( stat( file_path.c_str(), &st ) != 0 )
    {
        *error = -1;
        return 0;
    }

    std::FILE *fp = std::fopen( file_path.c_str(), "rb" );
    if ( fp == nullptr )
    {
        *error = -2;
        return 0;
    }

    // EEPROM - lesen
    if ( std::fread( m_eep_mem.data(), sizeof( uint8_t ), m_eep_mem.size(), fp ) != m_eep_mem.size() )
    {
        *error = -3;
        res = 0;
        goto error_label;
    }

    // Dateiheader lesen.
    if ( std::fread( m_file_header_mem.data(), sizeof( uint8_t ), m_file_header_mem.size(), fp ) != m_file_header_mem.size() )
    {
        *error = -3;
        res = 0;
        goto error_label;
    }

    // Headerdaten pruefen, ggf. Anpassungen durchfuehren.
    if ( !checkHeader() )
    {
        // Header hat CRC Fehler.
        *error = -4;
        res = 0;
        goto error_label;
    }

    if ( IsHeader() )
    {
        // Mindestversion des Studio pruefen.
        if ( !(flags & fiv_min_studio) && m_file_header.minStudioVersion )
        {
            SoftwarebuildVersion minVersion( m_file_header.minStudioVersion );
            if ( minVersion > m_studioVersion )
            {
                *error = -5;
                res = 0;
                goto error_label;
            }
        }
        // Sollteversion des Studio pruefen?
        if ( !(flags & fiv_should_studio) && !(m_file_header.flags & hfiv_studio) && m_file_header.shouldStudioVersion )
        {
            SoftwarebuildVersion shouldVersion( m_file_header.shouldStudioVersion );
            if ( shouldVersion > m_studioVersion )
            {
                *error = -6;
                res = 0;
                goto error_label;
            }
        }

        // Mindestversion der Library pruefen.
        if ( !(flags & fiv_min_library) && m_file_header.minLibraryVersion )
        {
            SoftwarebuildVersion minVersion( m_file_header.minLibraryVersion );
            if ( minVersion > m_libraryVersion )
            {
                *error = -7;
                res = 0;
                goto error_label;
            }
        }
        // Sollteversion der Library pruefen?
        if ( !(flags & fiv_should_library) && !(m_file_header.flags & hfiv_library) && m_file_header.shouldLibraryVersion )
        {
            SoftwarebuildVersion shouldVersion( m_file_header.shouldLibraryVersion );
            if ( shouldVersion > m_libraryVersion )
            {
                *error = -8;
                res = 0;
                goto error_label;
            }
        }
    }
    else
    {
        // Den Header im Standard wiederherstellen.
        resetFileHeader();
    }

    // Setupgroesse
    sizeSetup = endian_reverse( m_flash_header.flashSizeSUS );
    m_flashSizeSU2 = 0;
    // Ermittlung der Groesse des zu lesenden Speicherblocks fuer Flash.
    if ( m_flash_header.fileVersion == 0xDF02 && sizeSetup > FLASH_SIZE_SU )
    {
        m_flashSizeSU2 = FLASH_SIZE_SU2;
    }

    // Flashbereich - lesen
    if ( std::fread( m_flash_mem.data(), sizeof( uint8_t ), FLASH_SIZE_SU + m_flashSizeSU2, fp ) != (FLASH_SIZE_SU + m_flashSizeSU2) )
    {
        *error = -3;
        res = 0;
        goto error_label;
    }

    // ListFile - lesen
    for ( auto i = 0u; i < m_listFileName.size(); i++ )
    {
        // String inkl. Längenangabe aus Datei lesen.
        auto size = readCString( m_listFileName[i], fp );
        if ( size == 0 )
        {
            *error = -3;
            res = 0;
            goto error_label;
        }

        // Gesamtanzahl gelesener Zeichen halten.
        listFileSize += size;
    }

    // Konfiguration - lesen
    if ( std::fread( m_config_mem.data(), sizeof( uint8_t ), m_config_mem.size(), fp ) != m_config_mem.size() )
    {
        *error = -3;
        res = 0;
        goto error_label;
    }

    // Sind ab aktueller Postition noch 16kByte erweiterter Speicher vorhanden?
    if ( (uint32_t)st.st_size >=
        (m_eep_mem.size() + m_file_header_mem.size() + FLASH_SIZE_SU + m_flashSizeSU2 + listFileSize + m_config_mem.size()
          + m_extended_mem.size()) )
    {
        // Erweiterter Bereich - lesen: 16kByte Extended Flash
        if ( fread( m_extended_mem.data(), sizeof( uint8_t ), m_extended_mem.size(), fp ) != m_extended_mem.size() )
        {
            *error = -3;
            res = 0;
            goto error_label;
        }

        // existiert?
        if ( m_extended_header.check != EXTENDED_CHECK )
        {
            *error = -9;
            res = 0;
            goto error_label;
        }
    }
    else
    {
        *error = -9;
        res = 0;
        goto error_label;
    }

    // Sind noch ListFile fuer Zutrittskontrolle vorhanden?
    if ( (uint32_t)st.st_size >=
        (m_eep_mem.size() + m_file_header_mem.size() + FLASH_SIZE_SU + m_flashSizeSU2 + listFileSize + m_config_mem.size()
          + m_extended_mem.size() + m_listEntranceFileName.size()) )
    {
        // ListFile - lesen
        for ( auto i = 0u; i < m_listEntranceFileName.size(); i++ )
        {
            // String inkl. Längenangabe aus Datei lesen.
            auto size = readCString( m_listEntranceFileName[i], fp );
            if ( size == 0 )
            {
                *error = -3;
                res = 0;
                goto error_label;
            }

            // Gesamtanzahl gelesener Zeichen halten.
            entranceListFileSize += size;
        }

        // Loeschen, wenn ZK nicht aktiviert.
        if ( m_config.enableEntrance == 0 )
        {
            for ( auto &s : m_listEntranceFileName )
            {
                s.clear();
            }
            // CRC von Pruefung ausschliessen.
            m_file_header.crcListEntranceFile = 0;
            // CRC ungueltig.
            m_file_header.crc = 0;
        }
    }
    else
    {
        *error = -9;
        res = 0;
        goto error_label;
    }

    // Sind noch ListFile fuer Zutrittskontrolle 2 vorhanden?
    if ( (uint32_t)st.st_size >=
        (m_eep_mem.size() + m_file_header_mem.size() + FLASH_SIZE_SU + m_flashSizeSU2 + listFileSize + m_config_mem.size()
          + m_extended_mem.size() + entranceListFileSize + m_listEntrance2FileName.size()) )
    {
        // ListFile - lesen
        for ( auto i = 0u; i < m_listEntrance2FileName.size(); i++ )
        {
            // String inkl. Längenangabe aus Datei lesen.
            auto size = readCString( m_listEntrance2FileName[i], fp );
            if ( size == 0 )
            {
                *error = -3;
                res = 0;
                goto error_label;
            }

            // Gesamtanzahl gelesener Zeichen halten.
            entrance2ListFileSize += size;
        }

        // Loeschen, wenn ZK nicht aktiviert.
        if ( m_config.enableEntrance2 == 0 )
        {
            for ( auto &s : m_listEntrance2FileName )
            {
                s.clear();
            }
            // CRC von Pruefung ausschliessen.
            m_file_header.crcListEntrance2File = 0;
            // CRC ungueltig.
            m_file_header.crc = 0;
        }
    }
    else
    {
        /*
        *error = -9;
        res = 0;
        goto error_label;
        */
        // (BO)Bugfix: 2012_03_19
        // Diese Dateiangaben muessen noch nicht verfuegbar sein.
        // Es koennten Dateien fuer AEIII+ ueber die Library geladen werden!
    }

    // Sind noch ExtendedFilePath fuer Zusatzdateien vorhanden?
    if ( (uint32_t)st.st_size >=
        (m_eep_mem.size() + m_file_header_mem.size() + FLASH_SIZE_SU + m_flashSizeSU2 + listFileSize + m_config_mem.size()
          + m_extended_mem.size() + entranceListFileSize + entrance2ListFileSize + m_strExtendedFilePath.size()) )
    {
        // ListFile - lesen
        for ( auto i = 0u; i < m_strExtendedFilePath.size(); i++ )
        {
            // String inkl. Längenangabe aus Datei lesen.
            auto size = readCString( m_strExtendedFilePath[i], fp );
            if ( size == 0 )
            {
                *error = -3;
                res = 0;
                goto error_label;
            }

            // Gesamtanzahl gelesener Zeichen halten.
            extendedFilePathSize += size;
        }
    }
    else
    {
        // Diese Dateiangaben muessen noch nicht verfuegbar sein.
        // Es koennen Dateien ab DatafoxStudioIV - Version 04.01.01.10 geladen werden!
    }

    if ( IsHeader() )
    {
        // CRC der Dateibereiche pruefen
        if ( m_file_header.crcEep && m_file_header.crcEep != GetCrcEep() )
        {
            *error = -10;
            res = 0;
            goto error_label;
        }
        m_file_header.crcEep = 0;

        if ( m_file_header.crcEepExtended && m_file_header.crcEepExtended != GetCrcEepExtended() )
        {
            *error = -11;
            res = 0;
            goto error_label;
        }
        m_file_header.crcEepExtended = 0;

        if ( m_file_header.crcFlash && m_file_header.crcFlash != GetCrcFlash() )
        {
            *error = -12;
            res = 0;
            goto error_label;
        }
        m_file_header.crcFlash = 0;

        if ( m_file_header.crcListFile && m_file_header.crcListFile != GetCrcListFile() )
        {
            *error = -13;
            res = 0;
            goto error_label;
        }
        m_file_header.crcListFile = 0;

        if ( m_file_header.crcConfig && m_file_header.crcConfig != GetCrcConfig() )
        {
            *error = -14;
            res = 0;
            goto error_label;
        }
        m_file_header.crcConfig = 0;

        if ( m_file_header.crcFlashExtended && m_file_header.crcFlashExtended != GetCrcFlashExtended() )
        {
            *error = -15;
            res = 0;
            goto error_label;
        }
        m_file_header.crcFlashExtended = 0;

        if ( m_file_header.crcListEntranceFile && m_file_header.crcListEntranceFile != GetCrcListEntranceFile() )
        {
            *error = -16;
            res = 0;
            goto error_label;
        }
        m_file_header.crcListEntranceFile = 0;

        if ( m_file_header.crcListEntrance2File && m_file_header.crcListEntrance2File != GetCrcListEntrance2File() )
        {
            *error = -17;
            res = 0;
            goto error_label;
        }
        m_file_header.crcListEntrance2File = 0;

        if ( m_file_header.crcListExtendedFile && m_file_header.crcListExtendedFile != GetCrcListExtendedFile() )
        {
            *error = -18;
            res = 0;
            goto error_label;
        }
        m_file_header.crcListExtendedFile = 0;
    }

    // Gerätetypen als Setuptypen in einen einheitlichen Setuptypen vereinheitlichen.
    // #####
    {
        uint8_t setupType = static_cast<uint8_t>(m_flash_header.setupType);
        // EVO 3.5 Pure wird zu EVO 2.8 Pure, dies scheint es mal als eigenständigen Setuptypen gegeben zu haben!!
        if ( setupType == 24 )
        {
            m_flash_header.setupType = SetupType::EVO_2_8_Pure;
        }
    }

	m_extended_header = *(StructExtendedFlash *)m_flashExtended;

    // Eintragen in der Extended - Struktur
    if ( m_extended_header.offsetExtendedStruct )
    {
        // Die Werte fuer Dateinamen und Zeitstempel der Modifizierung eintragen.
        // Dateinamen extrahieren
        char *fName = (char *)&file_path[file_path.size() - 1];
        while ( fName > &file_path[0] )
        {
            // Maskierte Leerzeichen (unix)
            if ( *(fName - 1) == '\\' && *fName == ' ' )
            {

            }
            else if ( *(fName - 1) == '\\' )
            {
                // Pfadbestandteil vor Dateinamen (windows)
                break;
            }
            else if ( *(fName - 1) == '/' )
            {
                // Pfadbestandteil vor Dateinamen (unix)
                break;
            }

            fName--;
        }

        char mTimeStamp[20]; // 2009-05-19 09:16:22
        struct tm *ptm = localtime( &st.st_mtime );
        sprintf( mTimeStamp, "%02i%02i-%02i-%02i %02i:%02i:%02i",
                 std::min( 99, std::abs( (ptm->tm_year + 1900) / 100 ) ), std::min( 99, std::abs( (ptm->tm_year + 1900) % 100 ) ),
                 std::min( 12, std::abs( ptm->tm_mon + 1 ) ),
                 std::min( 31, std::abs( ptm->tm_mday ) ),
                 std::min( 23, std::abs( ptm->tm_hour ) ),
                 std::min( 59, std::abs( ptm->tm_min ) ),
                 std::min( 59, std::abs( ptm->tm_sec ) ) );

        StructExtended *extended = (StructExtended *)&m_flashExtended[m_extended_header.offsetExtendedStruct];
        strncpy( extended->fileName, fName, sizeof( extended->fileName ) );
        // Nullterminierung sicherstellen
        extended->fileName[sizeof( extended->fileName ) - 1] = 0;
        strcpy( extended->fileTimestamp, mTimeStamp );
    }

error_label:
    if ( fp )
    {
        std::fclose( fp );
    }

    return res;
}

int32_t FileAES::save( const std::string &file_path, int32_t *error )
{
    int32_t res = 1;
    uint32_t sizeSetup;

    *error = 0;
    std::FILE *fp = std::fopen( file_path.c_str(), "wb" );
    if ( fp == nullptr )
    {
        return 0;
    }

    // Groesse der Setupdaten eintragen
    sizeSetup = GetFlashSetupSize();
    m_flash_header.flashSizeSUS = endian_reverse( sizeSetup );

    // Ermittlung der Groesse des zu schreibenden Speicherblocks fuer Flash.
    if ( sizeSetup > FLASH_SIZE_SU )
    {
        m_flashSizeSU2 = FLASH_SIZE_SU2;
        m_flash_header.fileVersion = 0xDF02;
    }
    else
    {
        m_flashSizeSU2 = 0;
        m_flash_header.fileVersion = 0xDF01;
    }

    // EEPROM - schreiben
    if ( std::fwrite( m_eep_mem.data(), sizeof( uint8_t ), m_eep_mem.size(), fp ) != m_eep_mem.size() )
    {
        res = 0;
        goto error_label;
    }

    // Headerdaten erstellen, ggf. Anpassungen durchfuehren.
    if ( !buildHeader() )
    {
        res = 0;
        goto error_label;
    }

    // Dateiheaderbereich schreiben
    if ( std::fwrite( m_file_header_mem.data(), sizeof( uint8_t ), m_file_header_mem.size(), fp ) != m_file_header_mem.size() )
    {
        res = 0;
        goto error_label;
    }

    // Flashbereich - schreiben
    if ( std::fwrite( m_flash_mem.data(), sizeof( uint8_t ), FLASH_SIZE_SU + m_flashSizeSU2, fp ) != (FLASH_SIZE_SU + m_flashSizeSU2) )
    {
        res = 0;
        goto error_label;
    }

    // ListFile - schreiben
    for ( auto i = 0u; i < m_listFileName.size(); i++ )
    {
        auto size = writeCString( m_listFileName[i], fp );
        if ( size == 0 )
        {
            res = 0;
            goto error_label;
        }
    }

    // Konfiguration - schreiben
    if ( std::fwrite( m_config_mem.data(), sizeof( uint8_t ), m_config_mem.size(), fp ) != m_config_mem.size() )
    {
        res = 0;
        goto error_label;
    }

    // Erweiterter Speicher - schreiben
    if ( std::fwrite( m_extended_mem.data(), sizeof( uint8_t ), m_extended_mem.size(), fp ) != m_extended_mem.size() )
    {
        res = 0;
        goto error_label;
    }

    // Zutrittskontrolle - ListFile - schreiben
    for ( auto i = 0u; i < m_listEntranceFileName.size(); i++ )
    {
        auto size = writeCString( m_listEntranceFileName[i], fp );
        if ( size == 0 )
        {
            res = 0;
            goto error_label;
        }
    }
    // Zutrittskontrolle - ListFile - schreiben
    for ( auto i = 0u; i < m_listEntrance2FileName.size(); i++ )
    {
        auto size = writeCString( m_listEntrance2FileName[i], fp );
        if ( size == 0 )
        {
            res = 0;
            goto error_label;
        }
    }

    // Zusatzpfadangaben - schreiben
    for ( auto i = 0u; i < m_strExtendedFilePath.size(); i++ )
    {
        auto size = writeCString( m_strExtendedFilePath[i], fp );
        if ( size == 0 )
        {
            res = 0;
            goto error_label;
        }
    }

error_label:
    if ( fp )
    {
        std::fclose( fp );
    }

    return res;
}

//########################
// CRC der Dateibereiche
//########################

uint32_t FileAES::crcCString( const std::string &s, uint32_t crc, bool closure )
{
    // Nach MFC - CString Konvention Berechnung durchführen.
    if ( s.size() > 0xFF )
    {
        uint8_t len[3];
        len[0] = 0xFF;
        *(uint16_t *)&len[1] = (uint16_t)s.size();
        crc = GetCRC32( len, sizeof( len ), crc, false );
    }
    else
    {
        uint8_t len = (uint8_t)s.size();
        crc = GetCRC32( &len, sizeof( len ), crc, false );
    }

    // String rausschreiben
    return GetCRC32( (const uint8_t *)s.data(), (uint32_t)s.size(), crc, closure );
}

///< CRC ueber Eep-Bereich.
uint32_t FileAES::GetCrcEep()
{
    return GetCRC32( m_eep_mem.data(), (uint32_t)m_eep_mem.size() );
}

///< CRC ueber EepExtended-Bereich.
uint32_t FileAES::GetCrcEepExtended()
{
    return GetCRC32( &m_file_header_mem[m_file_header.bugfixSize], (uint32_t)m_file_header_mem.size() - m_file_header.bugfixSize );
}

///< CRC ueber Flash-Bereich.
uint32_t FileAES::GetCrcFlash()
{
    // Groesse der Setupdaten
    uint32_t sizeSetup = GetFlashSetupSize();

    // Ermittlung der Groesse des zu beachtenden Speicherblocks fuer Flash.
    if ( sizeSetup > FLASH_SIZE_SU )
    {
        m_flashSizeSU2 = FLASH_SIZE_SU2;
    }

    return GetCRC32( m_flash_mem.data(), FLASH_SIZE_SU + m_flashSizeSU2 );
}

///< CRC ueber die Listendateiangaben.
uint32_t FileAES::GetCrcListFile()
{
    uint32_t crc = 0;
    for ( auto i = 0u; i < m_listFileName.size(); i++ )
    {
        // CRC über die einzelnen Strings, so wie sie in die Datei geschrieben würden, berechnen.
        crc = crcCString( m_listFileName[i], crc, (i == (m_listFileName.size() - 1)) );
    }

    return crc;
}

///< CRC ueber die Config-Struktur.
uint32_t FileAES::GetCrcConfig()
{
    return GetCRC32( m_config_mem.data(), (uint32_t)m_config_mem.size() );
}

///< CRC ueber FlashExtended-Bereich.
uint32_t FileAES::GetCrcFlashExtended()
{
    return GetCRC32( m_extended_mem.data(), (uint32_t)m_extended_mem.size() );
}

///< CRC ueber die Listendateiangaben der ZK1.
uint32_t FileAES::GetCrcListEntranceFile()
{
    uint32_t crc = 0;
    for ( auto i = 0u; i < m_listEntranceFileName.size(); i++ )
    {
        // CRC über die einzelnen Strings, so wie sie in die Datei geschrieben würden, berechnen.
        crc = crcCString( m_listEntranceFileName[i], crc, (i == (m_listEntranceFileName.size() - 1)) );
    }

    return crc;
}

///< CRC ueber die Listendateiangaben der ZK2.
uint32_t FileAES::GetCrcListEntrance2File()
{
    uint32_t crc = 0;
    for ( auto i = 0u; i < m_listEntrance2FileName.size(); i++ )
    {
        // CRC über die einzelnen Strings, so wie sie in die Datei geschrieben würden, berechnen.
        crc = crcCString( m_listEntrance2FileName[i], crc, (i == (m_listEntrance2FileName.size() - 1)) );
    }

    return crc;
}

///< CRC ueber die Zusaetzlichen Listendateiangaben.
uint32_t FileAES::GetCrcListExtendedFile()
{
    uint32_t crc = 0;
    for ( auto i = 0u; i < m_strExtendedFilePath.size(); i++ )
    {
        // CRC über die einzelnen Strings, so wie sie in die Datei geschrieben würden, berechnen.
        crc = crcCString( m_strExtendedFilePath[i], crc, (i == (m_strExtendedFilePath.size() - 1)) );
    }

    return crc;
}

// Die zu verwendende Version der Datei
int32_t FileAES::NeededVersion()
{
    if ( GetFlashSetupSize() > FLASH_SIZE_SU )
    {
        return 2;
    }

    return 1;
}

uint16_t FileAES::endian_reverse( uint16_t val )
{
    return ((val & 0xFF) << 8)
        | ((val >> 8) & 0xFF);
}

uint32_t FileAES::endian_reverse( uint32_t val )
{
    return ((val & 0xFF) << 24)
        | ((val & 0xFF00) << 8)
        | ((val >> 8) & 0xFF00)
        | ((val >> 24) & 0xFF);
}

// Groesse der Setupdaten im Flash ermitteln
uint32_t FileAES::GetFlashSetupSize()
{
    // ##############################################################
    // Pointer sind direkte Adressen die im AEIII+:
    //   1. Im Motorolla - Format vorliegen.
    //   2. Mit einem Basisoffset -FLASH_KONFIG_START- berechnet sind.

    // Pointer auf ListenBeschreibungen ermitteln
    auto pLB = endian_reverse( m_flash_header.listStruct );
    pLB -= FLASH_KONFIG_START;

    // Pointerarray der Listenbeschreibungen durchschreiten
    int32_t i;
    for ( i = 0; i < MAX_RECORD_DESCRIPTION; i++ )
    {
        // Nullpointer suchen fuer Listenanzahl
        if ( *(uint32_t *)&m_flash[pLB + i * sizeof( uint32_t )] == 0 )
        {
            break;
        }
    }

    // Liste gefunden?
    if ( i > 0 )
    {
        // Pointer auf letzte definierte Liste
        pLB = endian_reverse( *(uint32_t *)&m_flash[pLB + (i - 1) * sizeof( uint32_t )] );
        pLB -= FLASH_KONFIG_START;

        StructRecordDescription *pLBStruct = (StructRecordDescription *)&m_flash[pLB];

        // Pointer auf Feldarray der Liste
        uint32_t pFeldArray = endian_reverse( pLBStruct->ptr );
        pFeldArray -= FLASH_KONFIG_START;

        // Felder vorhanden?
        if ( pLBStruct->fields )
        {
            // Pointer auf letztes Feld der Listenbeschreibung ermitteln.
            uint32_t pFeld = endian_reverse( *(uint32_t *)&m_flash[pFeldArray + (pLBStruct->fields - 1) * sizeof( uint32_t )] );
            pFeld -= FLASH_KONFIG_START;

            // Feldadresse + eigene Groesse bildet FlashSetupSize
            return pFeld + sizeof( StructRecordField );
        }


        // Keine Felder definiert!
        return pFeldArray;
    }

    // Keine Listenbeschreibung definiert!
    // Die Anzahl fuer Listenbeschreibungspointer muss mitberechnet werden,
    // um bei einer Uebertragung diese zu loeschen.
    return (pLB + MAX_RECORD_DESCRIPTION * sizeof( uint32_t ));
}

// Groesse des Erweiterten Setupbereichs ermitteln
uint32_t FileAES::GetExtendedFlashSetupSize()
{
    // Erweiterte Struktur vorhanden?
    if ( m_extended_header.check != EXTENDED_CHECK )
    {
        return 0;
    }

    return m_extended_header.offsetFreeScope;
}

// Groesse der Listen- oder Datensatzbeschreibungen im Flash ermitteln
// \param lists[in] Gibt an ob die Groesse der Listen- oder Datensatzbeschreibungen ermittelt werden sollen.
uint32_t FileAES::GetTablesSize( bool lists )
{
    // ##############################################################
    // Pointer sind direkte Adressen die im AEIII+:
    //   1. Im Motorolla - Format vorliegen.
    //   2. Mit einem Basisoffset -FLASH_KONFIG_START- berechnet sind.

    // Pointer auf Listen- oder DatensatzBeschreibungen ermitteln
    uint32_t pTables = endian_reverse( (lists) ? m_flash_header.listStruct : m_flash_header.dataStruct );
    pTables -= FLASH_KONFIG_START;

    // Pointerarray der Listenbeschreibungen durchschreiten
    int32_t i;
    for ( i = 0; i < MAX_RECORD_DESCRIPTION; i++ )
    {
        // Nullpointer suchen fuer letzte Tabelle
        if ( *(uint32_t *)&m_flash[pTables + i * sizeof( uint32_t )] == 0 )
        {
            break;
        }
    }

    // Tabelle gefunden?
    if ( i > 0 )
    {
        // Pointer auf die letzte definierte Tabelle
        pTables = endian_reverse( *(uint32_t *)(&m_flash[pTables + (i - 1) * sizeof( uint32_t )]) );
        pTables -= FLASH_KONFIG_START;

        StructRecordDescription *pTableStruct = (StructRecordDescription *)&m_flash[pTables];

        // Pointer auf Feldarray der Tabelle
        uint32_t pFeldArray = endian_reverse( pTableStruct->ptr );
        pFeldArray -= FLASH_KONFIG_START;

        // Felder vorhanden?
        if ( pTableStruct->fields )
        {
            // Pointer auf letztes Feld der Listenbeschreibung ermitteln.
            uint32_t pFeld = endian_reverse( *(uint32_t *)&m_flash[pFeldArray + (pTableStruct->fields - 1) * sizeof( uint32_t )] );
            pFeld -= FLASH_KONFIG_START;

            // Feldadresse + eigene Groesse bilden Endadresse
            return ((pFeld + sizeof( StructRecordField )) - pTables);
        }


        // Keine Felder definiert!
        return (pFeldArray - pTables);
    }

    // Keine Tabellen definiert!
    return 0;
}

// MD5 ueber den Dateiheader berechnen.
char *FileAES::GetMD5Header()
{
    if ( !IsHeader() )
    {
        return 0;
    }

    return MD5Bin( reinterpret_cast<uint8_t *>(&m_file_header), m_file_header.size );
}

// MD5 ueber Listen- oder Datensatzbeschreibungen im Flash berechnen.
// \param lists[in] Gibt an ob der MD5 der Listen- oder Datensatzbeschreibungen ermittelt werden soll.
char *FileAES::GetMD5Tables( bool lists )
{
    // ##############################################################
    // Pointer sind direkte Adressen die im AEIII+:
    //   1. Im Motorolla - Format vorliegen.
    //   2. Mit einem Basisoffset -FLASH_KONFIG_START- berechnet sind.

    // Pointer auf Listen- oder DatensatzBeschreibungen ermitteln
    uint32_t pTables = endian_reverse( (lists) ? m_flash_header.listStruct : m_flash_header.dataStruct );;
    pTables -= FLASH_KONFIG_START;

    // Pointerarray der Listenbeschreibungen durchschreiten
    int32_t size = GetTablesSize( lists );

    // Tabelle gefunden?
    if ( size > 0 )
    {
        return MD5Bin( &m_flash[pTables], size );
    }

    // Keine Tabellen definiert!
    return 0;
}

int32_t FileAES::ValidFirmwareVersion( SoftwarebuildVersion &fwVersion )
{
    if ( IsHeader() )
    {
        // Pruefung auf 04.01.xx
        if ( fwVersion >= "04.01.01.00" && fwVersion < "04.02.00.00" )
        {
            // Unsupported der Firmware 04.01.xx.yy pruefen.
            if ( is_invalid_header_version( m_file_header.minFirmware1Version ) )
            {
                return -1;
            }

            // Mindestversion der Firmware 04.01.xx.yy pruefen.
            if ( m_file_header.minFirmware1Version )
            {
                SoftwarebuildVersion minVersion( m_file_header.minFirmware1Version );
                if ( minVersion > fwVersion )
                {
                    return -2;
                }
            }
            // Sollteversion des der Firmware 04.01.xx.yy pruefen?
            if ( !(m_file_header.flags & hfiv_firmware1) && m_file_header.shouldFirmware1Version )
            {
                SoftwarebuildVersion shouldVersion( m_file_header.shouldFirmware1Version );
                if ( shouldVersion > fwVersion )
                {
                    return -3;
                }
            }
        }
        else if ( fwVersion >= "04.02.00.00" && fwVersion < "04.03.00.00" )
        {

            // Unsupported der Firmware 04.02.xx.yy pruefen.
            if ( is_invalid_header_version( m_file_header.minFirmware2Version ) )
            {
                return -4;
            }

            // Mindestversion der der Firmware 04.02.xx.yy pruefen.
            if ( m_file_header.minFirmware2Version )
            {
                SoftwarebuildVersion minVersion( m_file_header.minFirmware2Version );
                if ( minVersion > fwVersion )
                {
                    return -5;
                }
            }
            // Sollteversion der der Firmware 04.02.xx.yy pruefen?
            if ( !(m_file_header.flags & hfiv_firmware2) && m_file_header.shouldFirmware2Version )
            {
                SoftwarebuildVersion shouldVersion( m_file_header.shouldFirmware2Version );
                if ( shouldVersion > fwVersion )
                {
                    return -6;
                }
            }
        }
        else if ( fwVersion >= "04.03.00.00" && fwVersion < "04.04.00.00" )
        {
            // Unsupported der Firmware 04.03.xx.yy pruefen.
            if ( is_invalid_header_version( m_file_header.minFirmware3Version ) )
            {
                return -8;
            }

            // Mindestversion der der Firmware 04.03.xx.yy pruefen.
            if ( m_file_header.minFirmware3Version )
            {
                SoftwarebuildVersion minVersion( m_file_header.minFirmware3Version );
                if ( minVersion > fwVersion )
                {
                    return -9;
                }
            }
            // Sollteversion der der Firmware 04.03.xx.yy pruefen?
            if ( !(m_file_header.flags & hfiv_firmware3) && m_file_header.shouldFirmware3Version )
            {
                SoftwarebuildVersion shouldVersion( m_file_header.shouldFirmware3Version );
                if ( shouldVersion > fwVersion )
                {
                    return -10;
                }
            }
        }
        else
        {
            // Die Firmwareversion wird nicht unterstützt.
            return -7;
        }

        // Firmwareversion wird unterstuetzt.
        return 1;
    }

    // Es liegt kein Dateiheader vor!
    return 0;
}
